/**
 * Программа "счетчик секунд"
 */
// inslude the SPI library:
#include <SPI.h>


const int pin_OE  = 2;
const int pin_B  = 4;
const int pin_A  = 3;
const int pin_LD = 5;

// Буфер индикатора
unsigned long img_buf[ 16 ];

/**
 * Массив знакогенератора для цифр: 0 - 9
 */
const char digits[ 10 * 8 ] = 
{ //0
  0b00000,
  0b01100,
  0b10010,
  0b10010,
  0b10010,
  0b10010,
  0b01100,
  0b00000,
  //1 
  0b00000,
  0b00100,
  0b01100,
  0b00100,
  0b00100,
  0b00100,
  0b01110,
  0b00000,
  //2
  0b00000,
  0b01100,
  0b10010,
  0b00100,
  0b01000,
  0b10000,
  0b11110,
  0b00000,
  //3
  0b00000,
  0b11100,
  0b00010,
  0b01100,
  0b00010,
  0b00010,
  0b11100,
  0b00000,
  //4 
  0b00000,
  0b10010,
  0b10010,
  0b10010,
  0b11110,
  0b00010,
  0b00010,
  0b00000,
  //5
  0b00000,
  0b11110,
  0b10000,
  0b11100,
  0b00010,
  0b10010,
  0b01100,
  0b00000,
  //6
  0b00000,
  0b01110,
  0b10000,
  0b11100,
  0b10010,
  0b10010,
  0b01100,
  0b00000,
  // 7
  0b00000,
  0b11110,
  0b00010,
  0b00010,
  0b00100,
  0b01000,
  0b01000,
  0b00000,
  // 8
  0b00000,
  0b01100,
  0b10010,
  0b01100,
  0b10010,
  0b10010,
  0b01100,
  0b00000,
  // 9
  0b00000,
  0b01100,
  0b10010,
  0b10010,
  0b01110,
  0b00010,
  0b01100,
  0b00000,
};
/**
  * Аппаратные настройки
  */
void setup()
{
    pinMode (pin_A , OUTPUT);
    pinMode (pin_B , OUTPUT);
    pinMode (pin_OE , OUTPUT);
    pinMode (pin_LD, OUTPUT);
    // initialize SPI:
    SPI.begin(); 
    Serial.begin(115200);

}
/**
 * Вспомогательная функция: 
 * Вычисление значения data
 * @param row — номер строки
 * @param buf — указатель на буфер индикатора
 * @param regn — номер регистра индикатора
*/
int get_img_byte(unsigned long* buf, int col , int regn )
{
  int res = 0;
  int m = 24 - ( ( regn / 4 ) * 8 );
  int n = ( 15 - ( ( regn % 4 ) * 4 ) -  col );
  return buf[n]>> m;
  
}
/**
 * Переключение строк 
 * @param row — номер строки
 */
void set_row( int row )
{
  // Приводим row к нужному виду
  row &= 0x03;
  row = 3 - row;
  // Отключаем индикацию
  digitalWrite( pin_OE , LOW );
  // Переключаем ножк дешифратора
  digitalWrite( pin_A , ( row & 1 ) && HIGH );
  digitalWrite( pin_B , ( row & 2 ) && HIGH );
  // Выводим синхроимпульс регистра с параллельной загрузкой
  digitalWrite( pin_LD , HIGH );
  digitalWrite( pin_LD , LOW );
  // Включаем индикацию
  digitalWrite( pin_OE , HIGH );
}
/** 
 * Обновления индикатора - улучшеная версия
 * @param buf — указатель на буфер индикатора
 */
void upload( unsigned long*  buf )
{
  int regn;
  static int col = 0;
  if( col >= 4 )
      col = 0;
  
  for( regn = 0 ; regn < 16 ; regn++ )
      SPI.transfer( ~get_img_byte( buf , col , regn ) );     
  
  set_row( col );     
  ++col;
}
/**
 * Вывод цифры
 * @param x - координата x
 * @param y - координата y
 * @param dig - цифра
 */
void put_digit( int x , int y , int dig )
{
  for( int i = 0 ; i < 8 ; i++ )
  {
    if( ( ( i + y ) > 15 ) || ( ( x ) > 26 ) ||
        ( ( i + y ) < 0 ) || ( ( x ) < 0 ) )
      continue;
    unsigned long tmp = ( ( unsigned long ) digits[ dig * 8 + i ] << ( 26 - x ) );
    img_buf[ i + y ] |= tmp;
  }
}
/**
 * Вывод двузначного числа
 * @param x - координата x
 * @param y - координата y
 * @param dig - число
 */
void put_2dig( int x , int y , int d )
{
  put_digit( x + 5, y , d % 10 );
  d /= 10;
  put_digit( x , y , d % 10 );  
}
/**
 * Очистка буфера индикатора
 * @param buf — указатель на буфер индикатора
 */
void clear_buf( unsigned long*  buf )
{
  for( int i = 0 ; i < 16 ; i++ )
      buf[i] = 0;
}

/******************************************************/
void loop()
{
  setup();
  // счетчик циклов обновления 
  unsigned int counter = 0;
  // Счетчик 
  int cnty = 0;
  
  for( ; ; )
  {
    upload( img_buf );
    if ( counter % 100 == 0 )
    {
      clear_buf( img_buf );

      // Увеличения координаты вывода значения 10-ов мс 
      // формирование бегущей строки
      ++cnty;
      // Если выехали за край индикатора - начнаем сначала
      if ( cnty > ( 32 + 8 ) )
        cnty = 0;
      // Получаем заначение мс с момента старта системы
      unsigned long time = millis() / 10;
      // Вывод бегущей строкой 10-ов мс
      put_2dig( cnty - 8 , 8 , ( time % 100 ) );
      time /= 100;
      // Вывод десятков секунд
      put_2dig( 20, 0 , ( time  % 100 ) );
      time /= 100;
      // Вывод сотен секунд
      put_2dig( 10, 0 , ( time % 100 ) );
      time /= 100;
      // Вывод десятков тысяч секунд
      put_2dig( 0, 0 , ( time % 100 ) );
    }
    counter++;
//    delay(1);
  }
  
}
